package cmdutil

import (
	"fmt"
	"io"
	"os"
	"strings"

	"github.com/mattn/go-isatty"
	"github.com/pachyderm/pachyderm/v2/src/internal/errors"
	"github.com/pachyderm/pachyderm/v2/src/internal/serde"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
)

// Encoder creates an encoder that writes data structures to os.Stdout in the
// serialization format 'format'.  If the 'format' passed is unrecognized
// (currently, 'format' must be 'json' or 'yaml') then pachctl exits
// immediately.
func Encoder(format string, w io.Writer) serde.Encoder {
	format = strings.ToLower(format)
	if format == "" {
		format = "json"
	}
	e, err := serde.GetEncoder(format, w,
		serde.WithIndent(2),
		serde.WithOrigName(true),
	)
	if err != nil {
		ErrorAndExitf("%s", err.Error())
	}
	return e
}

func OutputFlags(raw *bool, output *string) *pflag.FlagSet {
	outputFlags := pflag.NewFlagSet("", pflag.ExitOnError)
	outputFlags.BoolVar(raw, "raw", false, "Disable pretty printing; serialize data structures to an encoding such as json or yaml")
	// --output is empty by default, so that we can print an error if a user
	// explicitly sets --output without --raw, but the effective default is set in
	// encode(), which assumes "json" if 'format' is empty.
	// Note: because of how spf13/flags works, no other StringVarP that sets
	// 'output' can have a default value either
	outputFlags.StringVarP(output, "output", "o", "", "Output format when --raw is set: \"json\" or \"yaml\" (default \"json\")")
	return outputFlags
}

func TimestampFlags(fullTimestamps *bool) *pflag.FlagSet {
	timestampFlags := pflag.NewFlagSet("", pflag.ContinueOnError)
	timestampFlags.BoolVar(fullTimestamps, "full-timestamps", false, "Return absolute timestamps (as opposed to the default, relative timestamps).")
	return timestampFlags
}

func PagerFlags(noPager *bool) *pflag.FlagSet {
	pagerFlags := pflag.NewFlagSet("", pflag.ContinueOnError)
	pagerFlags.BoolVar(noPager, "no-pager", false, "Don't pipe output into a pager (i.e. less).")
	return pagerFlags
}

func readLine(r io.Reader) (string, error) {
	// Read one byte at a time to avoid over-reading
	result := ""
	buf := make([]byte, 1)
	for string(buf) != "\n" {
		_, err := r.Read(buf)
		if err != nil {
			return "", errors.EnsureStack(err)
		}
		result += string(buf)
	}
	return result, nil
}

// InteractiveConfirm will ask the user to confirm an action on the command-line with a y/n response.
func InteractiveConfirm() (bool, error) {
	fmt.Fprintf(os.Stderr, "Are you sure you want to do this? (y/N): ")
	s, err := readLine(os.Stdin)
	if err != nil {
		return false, err
	}
	fmt.Printf("got confirm result: %s\n", s)
	if s[0] == 'y' || s[0] == 'Y' {
		return true, nil
	}
	return false, nil
}

// PrintStdinReminder will print a reminder if pachctl expects input on stdin, and stdin is a
// terminal (and not a pipe from another command or file).
func PrintStdinReminder() {
	if fd := os.Stdin.Fd(); isatty.IsTerminal(fd) || isatty.IsCygwinTerminal(fd) {
		fmt.Fprintln(os.Stderr, "Reading from stdin.")
	}
}

// FileMustExist is an argument that can be passed to cobra.Command.Args to ensure that the argument
// at index i is a file that exists.
func FileMustExist(i int) cobra.PositionalArgs {
	return func(cmd *cobra.Command, args []string) error {
		if i >= len(args) {
			return nil
		}
		if _, err := os.Stat(args[i]); err != nil {
			return errors.Wrap(err, "stat")
		}
		return nil
	}
}
